import json
import logging
from collections import OrderedDict

from .time_parser import elastic_time_parse
from .inferencial_parser import inferencial_parse
from .util import data_types, upcast_data_types, specified_type2lakeweed_type, get_array_inner_type


def data_string2type_value(src_str: str, specified_types={}, logger=logging.getLogger("lakeweed__clickhouse")) -> (tuple, tuple, list):
    """
    Convert string to python dict with data types for Clickhouse

    Arguments:
        src_str {str} -- Json, JsonLines, or CSV string.

    Keyword Arguments:
        logger -- logger object (default: {logging.getLogger("lakeweed__clickhouse")})

    Returns:
        tuple -- return tuple (columns, types, values_list).
            columns -- tuple of column name.
                       In Json, all nested kays(properties) are fltten.
                       In JsonLines, if JSONs has different keys each other, this function will return union set of keys.

            types -- A tuple of column type name for clickhouse.
                     Order of inside list values correspond to a columns order.

            values_list -- List of values tuple. Inside tuple correspoind to a record. For example, src string contains 2 records, length of Record List will be 2.
                           Order of inside list values correspond to a columns order. In JsonLines, if there is not key in some record, will be set None value.
    """

    (format, src_keys, src_values_list) = inferencial_parse(src_str, specified_types, "__", logger)

    # 1. Mapping type of value in src_values_list to lakeweed supporeted types.
    src_types_list = [data_types(values) for values in src_values_list]

    # 2. Upcasting types
    upcasted_types = upcast_data_types(src_types_list)

    # 3. Overwrite by specified_types if exists
    column_types = OrderedDict()
    for c, t in zip(src_keys, upcasted_types):
        column_types[c] = t

    for k, t in specified_types.items():
        if k not in column_types:
            continue
        column_types[k] = specified_type2lakeweed_type(t)

    # 4. [Depends on DBMS] If the type could not be estimated, it should be a String because there is no information to estimate it
    for c, t in column_types.items():
        if t is None:
            column_types[c] = 'String'
            continue
        inner_type = get_array_inner_type(t)
        if inner_type == 'Empty':
            column_types[c] = 'Array(String)'
        if inner_type == 'None':
            column_types[c] = 'Array(String)'


    # 5. [Depends on DBMS] Converting values according to the type and add columns if necessary
    ch_values_list = []
    ch_column_types = OrderedDict()
    for values in src_values_list:
        ch_values = []
        for c, t, v in zip(column_types.keys(), column_types.values(), values):
            ch_c, ch_t, ch_v = __data_value_specified(c, v, t)
            ch_column_types[ch_c] = ch_t
            ch_values.append(ch_v)
        ch_values_list.append(tuple(ch_values))

    columns_res = tuple(ch_column_types.keys())
    types_res = tuple(ch_column_types.values())

    return (columns_res, types_res, ch_values_list)


def __data_value_specified(column, value, specified_type, depth=0) -> tuple:
    t = str(specified_type).upper()

    if t.startswith("ARRAY"):
        return __list_data_value_specified(column, value, specified_type, depth)

    if t in ["FLOAT"]:
        v = convert_or_default(lambda: float(value), None)
        return (column, "Float64", v)
    if t in ["INT"]:
        v = convert_or_default(lambda: int(value), None)
        return (column, "Int64", v)
    if t in ['BOOL']:
        v = convert_or_default(lambda: 1 if bool(value) else 0, None)
        if value is None:
            v = None
        return (column, "UInt8", v)
    if t == 'DATETIME':
        dt, ns = convert_or_default(lambda: elastic_time_parse(str(value)).tupple(), (None, None))

        return (column, "DateTime64(6)", dt)
    if t in ['STRING', 'STR']:
        if isinstance(value, (list, dict, bool)):
            value = convert_or_default(lambda: json.dumps(value), None)
        v = convert_or_default(lambda: None if value is None else str(value), None)
        return (column, "String", v)
    # Todo need verify spec file method.
    logging.warning(f"Specified type '{str(specified_type)}' is not supported.")
    return (column, None, value)


def __list_data_value_specified(column, list_value, specified_type, depth) -> list:  # [(column, "Array(Type)", value), ]

    if depth > 0:
        json_str = convert_or_default(lambda: json.dumps(list_value), None)
        return (column, f"String", json_str)

    inner_type = get_array_inner_type(specified_type)

    values = []
    dummy, values_type, dummy = __data_value_specified("dummy", None, inner_type, depth + 1)

    if list_value is None or not isinstance(list_value, (list)):
        return (column, f"Array({values_type})", [])

    for v in list_value:
        col, typ, val = __data_value_specified("dummy", v, inner_type, depth + 1)
        values.append(val)
        values_type = typ

    return (column, f"Array({values_type})", values)


def convert_or_default(value_lambda, default):
    try:
        return value_lambda()
    except TypeError:
        return default
    except ValueError:
        return default



from dataclasses import dataclass

import datetime
from dateutil.parser import parse
from dateutil.tz import tzutc

import re

NanosecPattern = re.compile(r".+\.(\d+).*")
RequiredTimePattern = re.compile(r".*\d\d?[/:-]\d\d?.*")


@dataclass
class DateTimeWithNS:
    datetime: datetime.datetime
    nanosec: int
    original_string: str

    def tupple(self) -> (datetime.datetime, int):
        return (self.datetime, self.nanosec)

    def __str__(self) -> str:
        return self.original_string

    @classmethod
    def parse(cls, datetime_like_string: str):
        return elastic_time_parse(datetime_like_string)


def elastic_time_parse(src, logger=None) -> DateTimeWithNS:
    """Parse src string as datetime and nanosec part. Raise exception if src format is NOT valid. """
    if src is None:
        raise ValueError

    if not re.match(RequiredTimePattern, src):
        raise ValueError

    nano = 0
    ret = parse(src)
    if ret.tzinfo is None:
        ret = ret.replace(tzinfo=tzutc())

    m = NanosecPattern.match(src)
    if(m is not None):
        nano = int(m.group(1)[0:9].ljust(9, '0'))

    return DateTimeWithNS(ret, nano, src)



import json
from . import time_parser


def flatten(src: dict, target=None, prefix="", delimiter="."):
    if target is None:
        target = {}

    for k, v in src.items():
        key_name = f"{prefix}{k}"

        if type(v) is dict:
            if len(v) <= 0:
                target[key_name] = {}
            else:
                flatten(v, target=target, prefix=f"{key_name}{delimiter}", delimiter=delimiter)
        else:
            target[key_name] = v

    return target


def data_types(values) -> list:
    return [__data_type(v) for v in values]


def __data_type(value):
    if value is None:
        return None

    if type(value) is list:
        return __data_type_list(value)
    if type(value) in (float, int):
        return 'Float'
    if type(value) is bool:
        return 'Bool'
    # value is string
    try:
        time_parser.elastic_time_parse(str(value)).tupple()
        return 'DateTime'
    except ValueError:
        pass
    except Exception:
        pass  # Todo logging

    return 'String'


def __data_type_list(value_as_list):
    if len(value_as_list) <= 0:
        return 'Array(Empty)'

    # Todo: decide tye by upcast_data_types.
    inner_value = value_as_list[0]
    inner_type = __data_type(inner_value)
    return f"Array({inner_type})"


# -----------------------------------------------------------------
__TypeMap = {
    # sorted(["empty", "float", "bool", "datetime", "array(string)", "string"])
    # ['array(string)', 'bool', 'datetime', 'empty', 'float', 'string']

    # "Bool" + "Bool": "Bool",
    # "Bool" + "DateTime": "String",
    "Bool" + "Float": "Float",
    # "Bool" + "String": "String",

    # "DateTime" + "DateTime": "DateTime",
    # "DateTime" + "Float": "String",
    # "DateTime" + "String": "String",

    # "Float" + "Float": "Float",
    # "Float" + "String": "String",
    # "String" + "String": "String"
}

def upcast_data_types(types_list) -> list:
    if len(types_list) <= 0:
        return []

    import copy
    upcasted_type = copy.deepcopy(types_list[0])

    for types in types_list:
        upcasted_type = [__upcast_data_type(t1, t2) for t1, t2 in zip(upcasted_type, types)]

    return upcasted_type


def __upcast_data_type(type1, type2) -> str:
    if type1 is None:
        return type2
    if type2 is None:
        return type1
    if type1 == type2:
        return type1

    # If type1 and type2 are array both, upcast inner type.
    type1_inner = get_array_inner_type(type1)
    type2_inner = get_array_inner_type(type2)
    if None not in [type1_inner, type2_inner]:
        if type1_inner == "Empty":
            return type2
        if type2_inner == "Empty":
            return type1

        from_to = "".join(sorted([type1_inner, type2_inner]))
        inner_type = __TypeMap.get(from_to, "String")
        return f"Array({inner_type})"

    from_to = "".join(sorted([type1, type2]))
    return __TypeMap.get(from_to, "String")


# -----------------------------------------------------------------
__AliasTypes = {
    'FLOAT': "Float",
    'DOUBLE': "Float",
    'INT': "Int",
    'INTEGER': "Int",
    'BOOL': "Bool",
    'BOOLEAN': "Bool",
    'DATETIME': "DateTime",
    'STR': "String",
    'STRING': "String"
}


def specified_type2lakeweed_type(specified_type):
    return __AliasTypes.get(specified_type.upper(), specified_type)


# -----------------------------------------------------------------
def get_array_inner_type(array_type: str) -> str:
    import re
    pattern = re.compile(r"Array\((.+)\)", re.IGNORECASE)
    inner_type_m = pattern.match(array_type)

    if inner_type_m is None:
        return None

    return inner_type_m.group(1)
